/*
 * This file is part of JPuzzle.
 *
 * JPuzzle is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * JPuzzle is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with JPuzzle.  If not, see <http://www.gnu.org/licenses/>.
 *
 * JPuzzle Copyright © 2008 HackmanGT
 * HackmanGT@hotmail.com
 */

package JPuzzle.Graphics;

import JPuzzle.Event.MesaActionEvent;
import JPuzzle.Event.MesaActionListener;
import java.awt.Color;
import java.awt.Font;
import java.awt.Graphics;

/**
 * La mesa es la vista del tablero, se encarga de mostrar los datos en el
 * viewport, aceptar los eventos del mouse y reaccionar consecuentemente.
 * @author HackmanGT
 */
public class Mesa extends javax.swing.JPanel {

    private JPuzzle.Utilities.Tablero tablero;
    private boolean showNumbers = true;
    private boolean showImages = true;
    private boolean showGrid = true;
    private boolean animated = false;
    private boolean running = false;
    private int speed = 1000;
    private int count = 0;

    /**
     * Construye una nueva mesa descendiente de JPanel.
     */
    public Mesa() {
        initComponents();
    }

    /**
     * Inicializa el componente.
     */
    private void initComponents() {
        crearTablero(3, 3);
        addMouseListener(new java.awt.event.MouseAdapter() {
            @Override
            public void mouseClicked(java.awt.event.MouseEvent evt) {
                mouseClickedHandler(evt);
            }
        });
    }

    /**
     * Crea un nuevo tablero con el tamaño indicado en los parámetros.
     * @param ancho el nuevo ancho del tablero
     * @param alto el nuevo alto del tablero
     */
    private void crearTablero(int ancho, int alto) {
        tablero = new JPuzzle.Utilities.Tablero(ancho, alto);
        tablero.addChangeListener(new javax.swing.event.ChangeListener() {
            public void stateChanged(javax.swing.event.ChangeEvent evt) {
                stateChangedHandler(evt);
            }
        });
    }

    /**
     * Cuando sucede el evento <b>intercambiarCelda()</b> en la clase tablero
     * se genera una llamada a esta función por medio del listener.
     * @param e el evento recibido
     */
    private void stateChangedHandler(javax.swing.event.ChangeEvent evt) {
        repaint();
        count++;
        try {
            // Solamente desordenar y resolver pueden poner el valor true a
            // running; vamos a suponer que siempre que running sea verdadero
            // está corriendo en otro thread, y es ese el que duerme (sleep).
            if (running) {
                if (animated) {
                    firePropertyChange("cuenta", -1, count);
                    Thread.sleep(speed);
                }
            } else {
                firePropertyChange("cuenta", -1, count);
            }
        } catch (InterruptedException ex) {
            // Vamos a intentar interrumpir el proceso saliendo lo más pronto
            // posible del thread que ha sido interrumpido.
            animated = false;
            System.err.println(ex.getMessage());
        }
    }

    /**
     * Convierte las coordenadas absolutas del componente en coordenadas
     * relativas al tablero, intenta hacer el movimiento y verifica si
     * hay un ganador. Avisa enviando el evento propertyChange "winner".
     * @param evt el evento recibido
     */
    private void mouseClickedHandler(java.awt.event.MouseEvent evt) {
        if (!running) {
            int x = evt.getX() / (this.getWidth() / tablero.getAncho());
            int y = evt.getY() / (this.getHeight() / tablero.getAlto());
            if (tablero.intentarMover(new JPuzzle.Utilities.Posicion(x, y))) {
                if (tablero.verificarGanador()) {
                    firePropertyChange("winner", false, true);
                    count = 0;
                }
            }
        }
    }

    private static int rectX;
    private static int rectY;
    private static int rectW;
    private static int rectH;
    private static Font font;

    {
        try {
            font = new Font("Monospaced", Font.BOLD, 12);
        } catch (Exception e) {
            font = this.getFont();
        }
    }

    /**
     * Pinta el tablero, recorre la tabla dibujando la imágen, la rejilla y
     * el número correspondiente a cada celda. Calcula las coordenadas
     * correspondientes.
     * El acceso a la tabla lo realiza a traves de la función getCelda() en
     * la clase tablero; se espera que no exista ningún problema con la
     * sincronización (aunque no está sincronizado -- overhead), por eso se
     * almacena una referencia al tablero, por si este cambia.
     * Le falta optimización.
     * Debería de corren en el Swing Event Dispatching Thread.
     * @param g Graphics from Java System
     */
    @Override
    public void paint(Graphics g) {
        super.paint(g);

        JPuzzle.Utilities.Celda c = null;
        JPuzzle.Utilities.Tablero t = tablero;
        java.awt.image.BufferedImage b = null;

        int alto = t.getAlto();
        int ancho = t.getAncho();

        rectW = (this.getWidth() / ancho);
        rectH = (this.getHeight() / alto);

        g.setFont(font);
        g.setColor(isShowImages() ? Color.WHITE : Color.BLACK);

        for (int y = 0; y < alto; y++) {
            for (int x = 0; x < ancho; x++) {

                rectX = x * rectW;
                rectY = y * rectH;

                c = t.getCelda(x, y);
                if (c != null) {
                    b = c.getImagen();
                    if (showImages && b != null) {
                        g.drawImage(
                                b,
                                rectX, rectY,
                                rectW, rectH,
                                Color.BLACK,
                                null);
                    }
                    if (showGrid) {
                        g.draw3DRect(
                                rectX, rectY,
                                rectW, rectH,
                                true);
                    }
                    if (showNumbers) {
                        g.drawString(
                                String.valueOf(c.getValor()),
                                (rectX) + (rectW / 2),
                                (rectY) + (rectH / 2));
                    }
                } else {
                    if (!showImages) {
                        g.fill3DRect(
                                rectX +  8, rectY +  8,
                                rectW - 16, rectH - 16,
                                true);
                    }
                }
            }
        }
    }

    /**
     * Desordena la tabla, pero lo ejecuta en un thread separado,
     * para no afectar la velocidad de los eventos. Lo sincroniza
     * con el tablero.
     */
    public void desordenar() {
        new Thread() {
            @Override
            public void run() {
                synchronized (tablero) {
                    running = true;
                    fireAction(running, "Desordenar.");
                    tablero.desordenar(1000);
                    running = false;
                    fireAction(running, "");
                    firePropertyChange("cuenta", -1, count);
                    count = 0;
                }
            }
        }.start();
    }

    /**
     * Resuelve la tabla, pero lo ejecuta en un thread separado,
     * para no afectar la velocidad de los eventos. Lo sincroniza
     * con el tablero.
     */
    public void resolver() {
        new Thread() {
            @Override
            public void run() {
                synchronized (tablero) {
                    running = true;
                    fireAction(running, "Resolver.");
                    tablero.resolver();
                    running = false;
                    fireAction(running, "");
                    firePropertyChange("cuenta", -1, count);
                    count = 0;
                }
            }
        }.start();
    }

    /**
     * Cambia el tamaño del tablero a unas nuevas dimensiones.
     * Se sincroniza con el tablero anterior.
     * @param ancho el nuevo ancho
     * @param alto el nuevo alto
     */
    public void cambiarTamaño(int ancho, int alto) {
        running = false;
        synchronized (tablero) {
            crearTablero(ancho, alto);
            speed = (11 - ancho) * 125; // El tamaño máximo puede ser 10 x 10
            count = 0;
            firePropertyChange("cuenta", -1, count);
        }
        repaint();
    }

    /**
     * Agrega una vista receptora de eventos a la lista.
     * @param l la vista que va a recibir el evento
     */
    public void addMesaActionListener(MesaActionListener l) {
        listenerList.add(MesaActionListener.class, l);
    }

    /**
     * Remueve una vista receptora de eventos de la lista.
     * @param l la vista que se va a eliminar
     */
    public void removeMesaActionListener(MesaActionListener l) {
        listenerList.remove(MesaActionListener.class, l);
    }

    /**
     * Avisa a todas las vistas que hubo un cambio en el estado de "ejecutandose".
     */
    protected void fireAction(boolean running, String message) {
        Object[] listeners = listenerList.getListenerList();
        for (int i = listeners.length - 2; i >= 0; i -= 2) {
            if (listeners[i] == MesaActionListener.class) {
                MesaActionEvent mesaActionEvent = new MesaActionEvent(this, running, message);
                ((MesaActionListener) listeners[i + 1]).actionEvent(mesaActionEvent);
            }
        }
    }

    /**
     * Devuelve el estado de mostrando números.
     * @return el estado de mostrando números
     */
    public boolean isShowNumbers() {
        return showNumbers;
    }

    /**
     * Asigna el estado de mostrando números.
     * @param showNumbers el estado de mostrando números
     */
    public void setShowNumbers(boolean showNumbers) {
        this.showNumbers = showNumbers;
        repaint();
    }

    /**
     * Devuelve el estado de mostrando imágenes.
     * @return el estado de mostrando imágenes
     */
    public boolean isShowImages() {
        return showImages;
    }

    /**
     * Asigna el estado de mostrando imágenes.
     * @param showImages el estado de mostrando imágenes
     */
    public void setShowImages(boolean showImages) {
        this.showImages = showImages;
        repaint();
    }

    /**
     * Devuelve el estado de mostrando rejilla.
     * @return el estado de mostrando rejilla
     */
    public boolean isShowGrid() {
        return showGrid;
    }

    /**
     * Asigna el estado de mostrando rejilla.
     * @param showGrid el estado de mostrando rejilla
     */
    public void setShowGrid(boolean showGrid) {
        this.showGrid = showGrid;
        repaint();
    }

    /**
     * Devuelve el estado de está animado.
     * @return el estado de está animado
     */
    public boolean isAnimated() {
        return animated;
    }

    /**
     * Asigna el estado de está animado.
     * @param animated el estado de está animado
     */
    public void setAnimated(boolean animated) {
        this.animated = animated;
        repaint();
    }

    /**
     * Devuelve el estado de "ejecutandose".
     * @return el estado de "ejecutandose"
     */
    public boolean isRunning() {
        return running;
    }
}
