/*
 * This file is part of JPuzzle.
 *
 * JPuzzle is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * JPuzzle is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with JPuzzle.  If not, see <http://www.gnu.org/licenses/>.
 *
 * JPuzzle Copyright © 2008 HackmanGT
 * HackmanGT@hotmail.com
 */

package JPuzzle.Utilities;

/**
 * La clase Cerebro se encarga de resolver el tablero.
 * <b>Realmente esta clase se llama Pinky</b>.
 * No está optimizado para velocidad, se llaman demasiadas veces a la
 * funcion tablero.encontrarVacio(), así como tablero.encontrarCelda(),
 * cuando en algunos casos se podrían almacenar y pasar como parámetros.
 * <b>No es un cerebro muy inteligente, algunas veces hace movimientos que
 * obviamente no son necesarios; pero hacen mas sencillo el código.</b>
 * @author HackmanGT
 */
public class Cerebro {

    private Tablero tablero;

    /**
     * Construye un nuevo cerebro para resolver el tablero.
     * @param tablero el tablero a resolver
     */
    public Cerebro(Tablero tablero) {
        this.tablero = tablero;
    }

    /**
     * Resuelve el tablero, pasa una a una todas las celdas
     * intentando resolver cada una independientemente.
     */
    public void Resolver() {
        int tamano = tablero.getTamano() - 1;
        for (int pos = 1; pos < tamano; pos++) {
            if (!encontrado(pos)) {
                ResolverCelda(pos);
            }
        }
        if (!encontrado(tamano)) {
            intentarUltimoCambio(tamano);
        }
    }

    /**
     * Intenta resolver la posición de una celda indicada.
     * Si la celda está en la última columna, mueve el espacio vacío a la
     * primera columna de la fila inferior, mueve la celda superior hacia
     * abajo y todas las celdas de la fila superior un espacio a la izquierda,
     * al terminar vuelve a hacer la misma operación pero invertida.
     * @param pos la celda a resolver
     */
    private void ResolverCelda(int pos) {
        if (pos % tablero.getAncho() == 0) { // Cuando esté en la última columna
            int poy = (pos - 1) / tablero.getAncho();
            Posicion p = new Posicion(0, poy + 1);
            while (!encontradoVacioPosicion(p)) {
                while (encontrarMejorMovimiento(p, pos)) {
                }
                intentarUltimoCambio(pos);
            }
            MoverForzado(p, Posicion.Direccion.ARRIBA);
            for (int i = 0; i < tablero.getAncho() - 1; i++) {
                MoverForzado(p, Posicion.Direccion.DERECHA);
            }
        }

        while (!encontrado(pos)) { // Resulve la celda indicada
            Posicion p = encontrarMejorPosicion(pos);
            while (encontrarMejorMovimiento(p, pos)) {
            }
            intentarUltimoCambio(pos);
        }

        if (pos % tablero.getAncho() == 0) { // Cuando esté en la última columna
            int poy = (pos - 1) / tablero.getAncho();
            Posicion p = new Posicion(tablero.getAncho() - 2, poy);
            while (!encontradoVacioPosicion(p)) {
                while (encontrarMejorMovimiento(p, pos)) {
                }
            }
            for (int i = tablero.getAncho() - 2; i >= 0; i--) {
                MoverForzado(p, Posicion.Direccion.IZQUIERDA);
            }
            MoverForzado(p, Posicion.Direccion.ABAJO);
        }
    }

    /**
     * Valida que la posición indicada esté en el rango válido,
     * dentro de los límites de la tabla; y que la celda no esté fija,
     * que no sea un número que ya hallamos colocado.
     * @param p la posición a comprobar
     * @param pos el número de la celda que estamos trabajando
     * @return verdadero si es una posición válida, sino falso
     */
    private boolean esPosicionValida(Posicion p, int pos) {
        if (tablero.esPosicionValida(p)) {
            Celda c = tablero.getCelda(p.getX(), p.getY());
            if (c == null) {
                return true;
            } else {
                return c.getValor() > pos;
            }
        }
        return false;
    }

    /**
     * Verifica si la celda que estamos trabajando se encuentra posicionada
     * en el lugar final que le corresponde en la tabla.
     * @param pos la celda que estamos trabajando
     * @return verdadero si esta en su lugar correspondiente, sino falso
     */
    private boolean encontrado(int pos) {
        int poy = (pos - 1) / tablero.getAncho();
        Posicion q = tablero.encontrarCelda(pos);
        Posicion r = new Posicion((pos - 1) - (poy * tablero.getAncho()), poy);
        return q.equals(r);
    }

    /**
     * Verifica que la posición indicada sea igual a la posición del espacio
     * vacío en la tabla.
     * @param p la posición a verificar
     * @return verdadero si son iguales, sino falso
     */
    private boolean encontradoVacioPosicion(Posicion p) {
        Posicion q = tablero.encontrarVacio();
        return p.equals(q);
    }

    /**
     * Intenta encontrar la mejor posición para mover la celda vacía,
     * y poder cambiar la celda que estamos trabajando.
     * <br>
     * Busca la celda que estamos trabajando, así como calcula la posición
     * que debería tener en el tablero, compara las posiciónes arriba y abajo,
     * para ver cual es mejor, después compara izquierda y derecha, por último
     * compara las mejores posiciones entre las cuatro revisadas; para saber
     * cual está mas cerca para resolver el tablero.
     *
     * @param pos el número de la celda que estamos trabajando
     * @return la mejor posición encontrada, sino hay ninguna entonces nulo (realmente siempre hay una)
     */
    private Posicion encontrarMejorPosicion(int pos) {
        int poy = (pos - 1) / tablero.getAncho();

        Posicion q = tablero.encontrarCelda(pos);
        Posicion r = new Posicion((pos - 1) - (poy * tablero.getAncho()), poy);

        Posicion a = compararPosicion(r,
                q.relativa(Posicion.Direccion.ARRIBA),
                q.relativa(Posicion.Direccion.ABAJO),
                pos);
        Posicion b = compararPosicion(r,
                q.relativa(Posicion.Direccion.IZQUIERDA),
                q.relativa(Posicion.Direccion.DERECHA),
                pos);

        if (a != null && b != null) {
            return compararPosicion(r, a, b, pos);
        } else {
            if (a != null) {
                return a;
            } else {
                if (b != null) {
                    return b;
                }
            }
        }

        return null;
    }

    /**
     * Compara cual de las dos posiciones dadas <b>b</b> o <b>c</b> está
     * mas cercana a la posición <b>a</b>.
     * @param a la posición que es el objetivo
     * @param b la primera posición a comparar
     * @param c la segunda posicion a comparar
     * @param pos el número de la celda que estamos trabajando
     * @return la posición que está más cercana al objetivo, sino entonces nulo (siempre hay una)
     */
    private Posicion compararPosicion(Posicion a, Posicion b, Posicion c, int pos) {
        if (esPosicionValida(b, pos) && esPosicionValida(c, pos)) {
            if (Math.abs(a.getY() - b.getY()) < Math.abs(a.getY() - c.getY()) ||
                    Math.abs(a.getX() - b.getX()) < Math.abs(a.getX() - c.getX())) {
                return b;
            } else {
                return c;
            }
        } else {
            if (esPosicionValida(b, pos)) {
                return b;
            } else {
                if (esPosicionValida(c, pos)) {
                    return c;
                }
            }
        }
        return null;
    }

    /**
     * Este es el procedimiento que hace que se resuelva el tablero.
     * <br>
     * Búsca cual es el mejor movimiento en todas las direcciones
     * en busca de la posición p, para mover el espacio vacío en
     * esa dirección. Esta función se podría llamar "hacer uno o varios
     * movimientos en la dirección mas adecuada hacia la solución".
     *
     * @param p la posición a donde queremos llegar
     * @param pos el número de celda que estamos trabajando
     * @return verdadero si logró moverse en esa dirección, falso si necesita recalcular la mejor posición otra vez
     */
    private boolean encontrarMejorMovimiento(Posicion p, int pos) {
        Posicion q = tablero.encontrarVacio();
        Posicion t = null;

        if (q.equals(p)) {
            return false;
        }

        // Es muy importante que primero intente moverse en Y y después en X,
        // así como intente moverse ARRIBA antes de ABAJO e IZQUIERDA antes
        // de DERECHA.

        // Verifica arriba y abajo cual está mas cerca e intentar mover

        t = new Posicion(0, 0);
        if (cercanoY(q, p, t)) {
            if (intentarCambiar(q, t, pos)) {
                return true;
            }
        }

        // Verifica derecha e izquierda cual está mas cerca e intentar mover

        t = new Posicion(0, 0);
        if (cercanoX(q, p, t)) {
            if (intentarCambiar(q, t, pos)) {
                return true;
            }
        }

        Posicion r = tablero.encontrarCelda(pos);

        if (r.equals(p)) {
            return false;
        }

        // Si no pudo moverse antes, es por que está bloqueda la celda en el
        // lado izquierdo e intenta moverse en una forma de L hacia abajo y
        // la izquierda.

        if (q.getX() - 1 == r.getX()) {
            if (Mover(q, Posicion.Direccion.ABAJO, pos)) {
                if (Mover(q, Posicion.Direccion.IZQUIERDA, pos)) {
                    return true;
                }
            } else {

                // Si no pudo moverse antes, por que no pudo moverse hacia abajo
                // es por que estamos en la última linea, pero sigue bloqueda la
                // celda en el lado izquierdo, intenta un movimiento basado en un
                // truco, un movimiento específico que no modifica casí ninguna
                // celda importante o la regresa a su posición lo antes posible.

                if (q.getX() >= 2 && q.getX() <= tablero.getAncho() - 2) {
                    MoverForzado(q, Posicion.Direccion.IZQUIERDA);
                    MoverForzado(q, Posicion.Direccion.IZQUIERDA);
                    TrucoUltimasLineas(q);
                    return false;
                }
                if (q.getX() >= 2 && q.getX() <= tablero.getAncho() - 1) {
                    MoverForzado(q, Posicion.Direccion.IZQUIERDA);
                    MoverForzado(q, Posicion.Direccion.IZQUIERDA);
                    MoverForzado(q, Posicion.Direccion.IZQUIERDA);
                    TrucoUltimasLineas(q);
                    return false;
                }
            }
        }

        // Si no pudo moverse antes, es por que está bloqueda la celda en el
        // lado derecho e intenta moverse en una forma de L hacia abajo y
        // la derecha.

        if (q.getX() + 1 == r.getX()) {
            if (Mover(q, Posicion.Direccion.ABAJO, pos)) {
                if (Mover(q, Posicion.Direccion.DERECHA, pos)) {
                    return true;
                }
            } else {

                // Si no pudo moverse antes, por que no pudo moverse hacia abajo
                // es por que estamos en la última linea, pero sigue bloqueda la
                // celda en el lado derecho, intenta un movimiento basado en un
                // truco, un movimiento específico que no modifica casí ninguna
                // celda importante o la regresa a su posición lo antes posible.
                // El segundo truco creo que es evaluado solo para 3 x 3.

                if (q.getX() <= tablero.getAncho() - 4) {
                    TrucoUltimasLineas(q);
                    return false;
                }
                if (q.getX() + 1 == r.getX()) {
                    MoverForzado(q, Posicion.Direccion.ARRIBA);
                    MoverForzado(q, Posicion.Direccion.DERECHA);
                    MoverForzado(q, Posicion.Direccion.ABAJO);
                    MoverForzado(q, Posicion.Direccion.DERECHA);
                    MoverForzado(q, Posicion.Direccion.ARRIBA);
                    MoverForzado(q, Posicion.Direccion.IZQUIERDA);
                    MoverForzado(q, Posicion.Direccion.IZQUIERDA);
                    MoverForzado(q, Posicion.Direccion.ABAJO);
                    return false;
                }
            }
        }

        // Si llegó hasta aquí y no pudo moverse antes, es por que está bloqueda
        // la celda en el la parte superior e intenta moverse en una forma de L
        // hacia la derecha o izquierda y hacia arriba.

        // Es muy importante que en este caso sea primero a la DERECHA y luego
        // a la IZQUIERDA. Es decir, para saltarse una celda llena, la rodea
        // haciendo una U, primero con una L en los dos casos anteriores, y
        // despues una L invertida con estas líneas de código.

        if (q.getY() - 1 == r.getY()) {
            if (Mover(q, Posicion.Direccion.DERECHA, pos)) {
                if (Mover(q, Posicion.Direccion.ARRIBA, pos)) {
                    return true;
                }
            } else {
                if (Mover(q, Posicion.Direccion.IZQUIERDA, pos)) {
                    if (Mover(q, Posicion.Direccion.ARRIBA, pos)) {
                        return true;
                    }
                }
            }
        }

        return false;
    }

    /**
     * Comprueba cual posición es mas grande o más pequeña en Y, para modificar
     * el transformador en <b>c</b>; para saber si debemos movernos hacia
     * arriba o hacia abajo.
     * @param a la posición a donde deseamos avanzar
     * @param b la posición donde estamos
     * @param c la posición que funciona como transformador
     * @return falso si las dos son iguales, sino verdadero
     */
    private boolean cercanoY(Posicion a, Posicion b, Posicion c) {
        if (a.getY() > b.getY()) {
            c.setY(-1);
            return true;
        } else {
            if (a.getY() < b.getY()) {
                c.setY(1);
                return true;
            }

        }
        return false;
    }

    /**
     * Comprueba cual posición es mas grande o más pequeña en X, para modificar
     * el transformador en <b>c</b>; para saber si debemos movernos a la
     * derecha o a la izquierda.
     * @param a la posición a donde deseamos avanzar
     * @param b la posición donde estamos
     * @param c la posición que funciona como transformador
     * @return falso si las dos son iguales, sino verdadero
     */
    private boolean cercanoX(Posicion a, Posicion b, Posicion c) {
        if (a.getX() > b.getX()) {
            c.setX(-1);
            return true;
        } else {
            if (a.getX() < b.getX()) {
                c.setX(1);
                return true;
            }

        }
        return false;
    }

    /**
     * Mueve la posición indicada en los parámetros en la dirección indicada.
     * Nota: la función se podría optimizar extrayendo los if's del case y
     * realizando el case dentro de un solo if pero no resulta tan clara.
     * @param p la posición a mover
     * @param d la dirección a mover
     * @param pos el número de celda que estamos trabajando
     * @return verdadero si lo pudo mover, sino falso
     */
    private boolean Mover(Posicion p, Posicion.Direccion d, int pos) {
        switch (d) {
            case ARRIBA:
                if (intentarCambiar(p, Posicion.posicionDireccion(d), pos)) {
                    p.setY(p.getY() - 1);
                    return true;
                }
                break;
            case ABAJO:
                if (intentarCambiar(p, Posicion.posicionDireccion(d), pos)) {
                    p.setY(p.getY() + 1);
                    return true;
                }
                break;
            case DERECHA:
                if (intentarCambiar(p, Posicion.posicionDireccion(d), pos)) {
                    p.setX(p.getX() + 1);
                    return true;
                }
                break;
            case IZQUIERDA:
                if (intentarCambiar(p, Posicion.posicionDireccion(d), pos)) {
                    p.setX(p.getX() - 1);
                    return true;
                }
                break;
        }
        return false;
    }

    /**
     * Mueve la posición indicada en los parámetros en la dirección indicada
     * a la fuerza; eso significa que no verifica si se desplaza una celda
     * ya fija, que está en su posición (supuestamente) final, existe por
     * que se necesitan mover celdas y regresarlas manualmente.
     * Ver <b>TrucoUltimasLineas()</b>
     * @param p la posición a mover
     * @param d la dirección a mover
     * @return verdadero si lo pudo mover, sino falso
     */
    private boolean MoverForzado(Posicion p, Posicion.Direccion d) {
        switch (d) {
            case ARRIBA:
                if (intentarCambiarForzado(p, new Posicion(0, -1))) {
                    p.setY(p.getY() - 1);
                    return true;
                }
                break;
            case ABAJO:
                if (intentarCambiarForzado(p, new Posicion(0, 1))) {
                    p.setY(p.getY() + 1);
                    return true;
                }
                break;
            case DERECHA:
                if (intentarCambiarForzado(p, new Posicion(1, 0))) {
                    p.setX(p.getX() + 1);
                    return true;
                }
                break;
            case IZQUIERDA:
                if (intentarCambiarForzado(p, new Posicion(-1, 0))) {
                    p.setX(p.getX() - 1);
                    return true;
                }
                break;
        }
        return false;
    }

    /**
     * El truco consiste en realizar los pasos indicados en el orden correcto
     * para mover la celda dos posiciónes adelante sin modificar ninguna celda
     * en la línea superior ni inferior (realmente son las últimas lineas, no
     * hay línea inferior).
     * @param p la posición donde comienza el truco
     */
    private void TrucoUltimasLineas(Posicion p) {
        MoverForzado(p, Posicion.Direccion.ARRIBA);
        MoverForzado(p, Posicion.Direccion.DERECHA);
        MoverForzado(p, Posicion.Direccion.ABAJO);
        MoverForzado(p, Posicion.Direccion.DERECHA);
        MoverForzado(p, Posicion.Direccion.DERECHA);
        MoverForzado(p, Posicion.Direccion.ARRIBA);
        MoverForzado(p, Posicion.Direccion.IZQUIERDA);
        MoverForzado(p, Posicion.Direccion.IZQUIERDA);
        MoverForzado(p, Posicion.Direccion.IZQUIERDA);
        MoverForzado(p, Posicion.Direccion.ABAJO);
        MoverForzado(p, Posicion.Direccion.DERECHA);
        MoverForzado(p, Posicion.Direccion.DERECHA);
        MoverForzado(p, Posicion.Direccion.ARRIBA);
        MoverForzado(p, Posicion.Direccion.DERECHA);
        MoverForzado(p, Posicion.Direccion.ABAJO);
    }

    /**
     * Intenta cambiar la celda dada en la posición <b>a</b> transformada con
     * la posición (transformador) <b>b</b>; pos sirve para pasarlo a la validación.
     * @param a la posición de la celda que queremos probar a cambiar
     * @param b el transformador de la posición que queremos probar a cambiar
     * @param pos el número de la celda que estamos trabajando
     * @return verdadero si lo pudo cambiar, sino falso
     */
    private boolean intentarCambiar(Posicion a, Posicion b, int pos) {
        Posicion c = a.transformar(b);
        if (esPosicionValida(c, pos)) {
            tablero.intercambiarCelda(a, c);
            return true;
        }

        return false;
    }

    /**
     * Intenta cambiar la celda dada en la posición <b>a</b> transformada con
     * la posición (transformador) <b>b</b>. Forzado significa que no importa
     * que la celda ya este en su lugar, por que en algún momento la vamos a
     * regresar manualmente.
     * Ver <b>TrucoUltimasLineas()</b>
     * @param a la posición de la celda que queremos probar a cambiar
     * @param b el transformador de la posición que queremos probar a cambiar
     * @return verdadero si lo pudo cambiar, sino falso
     */
    private boolean intentarCambiarForzado(Posicion a, Posicion b) {
        Posicion c = a.transformar(b);
        if (tablero.esPosicionValida(c)) {
            tablero.intercambiarCelda(a, c);
            return true;
        }

        return false;
    }

    /**
     * Intenta hacer el último cambio; es decir, cambiar la celda que estamos
     * trabajando por la posición vacía, que ya debe de estar cercana, a una
     * celda de distancia.
     * @param pos la celda que estamos trabajando
     * @return verdadero si la pudo cambiar, sino falso
     */
    private boolean intentarUltimoCambio(int pos) {
        Posicion p = tablero.encontrarVacio();
        Posicion q = tablero.encontrarCelda(pos);

        if ((Math.abs(p.getX() - q.getX()) == 0 && Math.abs(p.getY() - q.getY()) == 1) ||
                (Math.abs(p.getX() - q.getX()) == 1 && Math.abs(p.getY() - q.getY()) == 0)) {
            tablero.intercambiarCelda(p, q);
            return true;
        }

        return false;
    }
}
