/*
 * This file is part of JPuzzle.
 *
 * JPuzzle is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * JPuzzle is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with JPuzzle.  If not, see <http://www.gnu.org/licenses/>.
 *
 * JPuzzle Copyright © 2008 HackmanGT
 * HackmanGT@hotmail.com
 */

package JPuzzle.Utilities;

/**
 * Las posiciones representan un par de coordenadas, sirven para almacenar,
 * comparar y realizar operaciones matemáticas (transformaciones) con
 * direcciones y coordenadas. Ver <b>java.awt.Point</b>
 * @author HackmanGT
 */
public class Posicion {

    /**
     * Enumeración de las posibles direcciones que sirven para transformar una
     * posición, permite generar una dirección aleatoria. El orden es importante
     * para la velocidad en la búsqueda de los movimientos.
     */
    protected enum Direccion {

        ARRIBA, ABAJO, IZQUIERDA, DERECHA;

        /**
         * Genera una dirección aleatoria.
         * @return una dirección aleatoria
         */
        public static Direccion random() {
            switch ((int) ((java.lang.Math.random() * 4) + 1)) {
                case 1:
                    return ARRIBA;
                case 2:
                    return ABAJO;
                case 3:
                    return DERECHA;
                case 4:
                    return IZQUIERDA;
                default:
                    throw new IndexOutOfBoundsException("Fuera del rango aleatorio.");
            }
        }
    }

    private int x;
    private int y;

    /**
     * Construye una nueva posición con el valor por defecto igual a 0, 0.
     */
    public Posicion() {
        initComponents(0, 0);
    }

    /**
     * Construye una nueva posición con los valores indicados en los
     * parámetros x, y.
     * @param x el valor inicial en el eje x
     * @param y el valor inicial en el eje y
     */
    public Posicion(int x, int y) {
        initComponents(x, y);
    }

    /**
     * Inicializa el componente.
     * @param x el valor inicial en el eje x
     * @param y el valor inicial en el eje y
     */
    private void initComponents(int x, int y) {
        this.setX(x);
        this.setY(y);
    }

    /**
     * Compara con otro objeto para revisar si es una posición, y si es igual
     * en los ejes x, y.
     * @param obj el objeto a comparar
     * @return verdadero si es una posición igual en los ejes x, y; sino falso
     */
    @Override
    public boolean equals(Object obj) {
        if (obj == null) {
            return false;
        }
        if (getClass() != obj.getClass()) {
            return false;
        }
        final Posicion other = (Posicion) obj;
        if (this.x != other.x) {
            return false;
        }
        if (this.y != other.y) {
            return false;
        }
        return true;
    }

    /**
     * Genera un valor hash para usar este componente en algoritmos de
     * ordenación y comparación.
     * @return un valor hash
     */
    @Override
    public int hashCode() {
        int hash = 3;
        hash = 41 * hash + this.x;
        hash = 41 * hash + this.y;
        return hash;
    }

    /**
     * Obtiene una nueva posición de transformación, en función de la dirección
     * indicada, con las coordenadas para realizar una tranformación.
     * @param d la dirección para obtener la nueva posición de transformación
     * @return una nueva posición de transformación con las coordenadas correctas
     */
    protected static Posicion posicionDireccion(Direccion d) {
        switch (d) {
            case ARRIBA:
                return new Posicion(0, -1);
            case ABAJO:
                return new Posicion(0, 1);
            case DERECHA:
                return new Posicion(1, 0);
            case IZQUIERDA:
                return new Posicion(-1, 0);
            default:
                throw new IndexOutOfBoundsException("Fuera del rango aceptado.");
        }
    }

    /**
     * Realiza una transformación de la posición actual en base a una posición
     * indicada, obteniendo como resultado la suma. (x + dx, y + dy). Si la
     * posición indicada es negativa regresa la operación inversa (x + (-dx)).
     * @param p la posición con la cual vamos a transformar
     * @return una nueva posición en función de la operación realizada
     */
    protected Posicion transformar(Posicion p) {
        return new Posicion(this.getX() + p.getX(), this.getY() + p.getY());
    }

    /**
     * Realiza una transformación relativa de la posición actual en base a una
     * dirección indicada, obteniendo como resultado una nueva posición.
     * @param d la direccion con la cual vamos a transformar la posición
     * @return una nueva posición en función de la dirección indicada
     */
    public Posicion relativa(Direccion d) {
        return transformar(posicionDireccion(d));
    }

    /**
     * Devuelve la coordenada en x.
     * @return la coordenada en x
     */
    public int getX() {
        return x;
    }

    /**
     * Asigna un valor a la coordenada en x.
     * @param x el valor de la coordenada en x
     */
    public void setX(int x) {
        this.x = x;
    }

    /**
     * Devuelve la coordenada en y.
     * @return la coordenada en y
     */
    public int getY() {
        return y;
    }

    /**
     * Asigna un valor a la coordenada en y.
     * @param y el valor de la coordenada en y
     */
    public void setY(int y) {
        this.y = y;
    }
}
